/*

  RT-11 file system module for Linux.

  By John Wilson <wilson@dbit.com>.

  Copyright (C) 1999-2000 by Digby's Bitpile, Inc.  All rights reserved.
  This program may be freely distributed provided that no charge is made for
  distribution and that all versions include this copyright notice and give
  proper credit to the original author.

  02/23/1999	JMBW	Created.
  10/13/2000	JMBW	Updated for 2.4.0-test9 kernel.
  10/14/2000	JMBW	Can get dir listing.
  10/15/2000	JMBW	Reading works, and .DSKs as subdirectories.
  10/17/2000	JMBW	Trims trailing NULs from file lengths.
  10/18/2000	JMBW	V1.0 (read-only version).

*/

/* le16_to_cpu and cpu_to_le16 are macros for byte-swapping */
/* in case we need them (might be better to do that by hand) */

#define MODULE
#ifndef __KERNEL__
#define __KERNEL__ 1	/* needed to enable FS defs in fs.h */
#endif
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>

/* RT-11 uses wall clock time so we need to know how to convert to UTC -- */
/* if your terminal isn't in the same time zone as the CPU, well, sorry */
extern struct timezone sys_tz;

#define RTMAGIC 0x31315452UL	/* 'RT11', little-endian, why not */
#define RTBLKSIZE 512		/* block size is always 512 (DEC hardware) */
#define RTBLKMASK 0777		/* mask for offset within block */
#define RTBLKBITS 9		/* log2(blocksize) */

typedef unsigned short word;	/* hopefully not much over 16 bits */
				/* (but certainly not under either) */

/* struct containing information needed for a directory lookup */
struct rt11dirbuf {
	kdev_t dev;		/* device record for current volume */
	unsigned int start;	/* starting blk # of logical disk (0=root) */
	unsigned int extra;	/* # of extra user-defined bytes per entry */
	word buf[7];		/* latest 7-word RT-11 dir entry */
				/* (also used for 5-word dir seg header) */
	struct buffer_head *bh;	/* buf descriptor for current dir block */
				/* (NULL if none allocated) */
	unsigned int dirseg;	/* current dir segment */
	unsigned int nextseg;	/* next dir segment in chain */
	unsigned long segmask;	/* bitmap of visited segments (LSB=1, b1=2) */
				/* (for detecting loops) */
	unsigned int offset;	/* offset into current seg, even # 0::1022 */
	unsigned int data;	/* starting data block for this file */
	unsigned int size;	/* size of last file in blocks */
};

/* return RADIX-50 code for a character */
static int radchar(char c)
{
	unsigned int n;

	/* don't check for $, ., or %, since they aren't allowed in */
	/* filename elements anyway */

	/* number */
	n=c-'0';
	if(n<=9) {
		n+=30;		/* digits start at 30 */
		goto done;
	}

	/* upper case letter */
	n=c-'A';
	if(n<26) {
		n++;		/* letters start at 1 */
		goto done;
	}

	/* lower case letter */
	n=c-'a';
	if(n<26) {
		n++;
		goto done;
	}

	n=0;			/* not a RAD50 character */

done:	return(n);
}

/* convert 3 chars from ASCII to RAD50 */
static unsigned int asc2rad(char **p,int *n)
{
	int a=0, b=0, c=0;

	if(*n==0) goto done;	/* nothing to get */
	a=radchar(**p);		/* grab a letter */
	if(a==0) goto done;	/* not RAD50, give up */
	(*p)++, (*n)--;		/* eat it */

	if(*n==0) goto done;	/* same for letter #2 */
	b=radchar(**p);
	if(b==0) goto done;
	(*p)++, (*n)--;

	if(*n==0) goto done;	/* and letter #3 */
	c=radchar(**p);
	if(c==0) goto done;
	(*p)++, (*n)--;

done:	return((a*050+b)*050+c);  /* compose 16-bit RAD50 word */
}

/* convert RAD50 word to 0-3 chars of ASCII */
static void rad2asc(word val,char **p,int *n)
{
	/* RADIX-50 lookup table (: is for invalid numbers >= 64000.) */
	/* we return lower-case so as not to seem jarring to Weenix Unies */
	/* (either case is fine when creating files) */
	static char rad50[] = "\0abcdefghijklmnopqrstuvwxyz$.%0123456789:";
	char a, b, c;

	/* split out RAD50 digits into three ASCII characters a/b/c */
	c=rad50[val%050];
	b=rad50[(val/050)%050];
	a=rad50[val/(050*050)];

	/* store first char if non-null */
	if(a!='\0') {
		**p=a;
		(*p)++, (*n)++;
	}

	/* same for other two */
	if(b!='\0') {
		**p=b;
		(*p)++, (*n)++;
	}

	if(c!='\0') {
		**p=c;
		(*p)++, (*n)++;
	}
}

/* read specified # of contiguous words from dir seg */
/* this would have been easy except that dir segs are two blocks long, and */
/* bread() currently works only with single 512-byte blocks, so the two */
/* halves of the dir seg are in two different buffers */

/* d	rt11dirbuf struct for this dir search */
/* p	pointer to where to store the 16-bit words */
/* n	number of words to transfer */
static int dirwords(struct rt11dirbuf *d,word *p,int n)
{
	register unsigned int relblk=0;	/* rel blk # in segment, 0 or 1 */
	int rc=-1;		/* return code, assume failure */

	while(n) {
		relblk+=(d->offset>>RTBLKBITS);	/* normalize address */
		d->offset&=RTBLKMASK;
		if(relblk>=2) goto punt;  /* must be inside the 2 blocks */

		/* if starting a new block, read it */
		if(d->offset==0) {
			/* flush old block, if any */
			if(d->bh!=NULL) brelse(d->bh);

			/* read new block */
			/* dir segs are 2 blks each starting @ blk 6 */
			d->bh=bread(d->dev,d->start+6+(d->dirseg-1)*2+relblk,
				RTBLKSIZE);

			if(d->bh==NULL) goto punt;  /* error */
		}

		/* copy words until done or at end of block */
		while(n) {
			/* assemble a little-endian 16-bit word */
			*p++=((unsigned char *)d->bh->b_data)[d->offset]|
				(((unsigned char *)d->bh->b_data)[d->offset+1]
				<<8);
			d->offset+=2, n--;  /* skip the 16-bit word */
			if(d->offset==RTBLKSIZE)
				break;	/* ran off end */
		}
	}
	rc=0;		/* success */

punt:	return(rc);
}

/* init for directory search */
static void rt11dirinit(struct rt11dirbuf *d,kdev_t dev,unsigned int start)
{
	d->dev=dev;	/* copy device descriptor pointer */
	d->start=start;	/* copy start of subdisk (0 if whole disk) */
	d->bh=NULL;	/* no buf descriptor yet */
	d->offset=0;	/* ready to start a fresh dir seg */
	d->dirseg=1;	/* starting dir segment */
	d->segmask=0UL;	/* none visited yet */
}

/* fetch next directory entry */
static int rt11dirent(struct rt11dirbuf *d)
{
next:	/* if at beginning of dir segment, process header */
	if(d->offset==0) {
		/* see if we've been here before (endless loop in dir) */
		if(d->segmask&(1UL<<(d->dirseg-1)))
			goto flush;

		/* nope, but mark it for next time */
		d->segmask|=(1UL<<(d->dirseg-1));

		/* fetch 5-word dir segment header */
		if(dirwords(d,d->buf,5)<0)
			goto flush;

		d->nextseg=d->buf[1];  /* save link to next seg */

		/* fetch and check # of extra user-defined bytes per dir */
		/* entry, if this is the first segment */
		if(d->dirseg==1) {
			d->extra=d->buf[3];

			/* must be even, and not too large to fit at least */
			/* the five-word dir seg header, one (7+EXTRA/2) */
			/* word dir entry, and the one-word dir terminator, */
			/* in a single dir seg */
			if((d->extra&1)||(d->extra>(1024-(5+7+1)*2)))
				goto flush;
		}

		/* get starting blk # of file data covered by this segment */
		d->data=d->start+d->buf[4];

		d->size=0;	/* pretend previous file was 0 blocks */
	}

	/* fetch next dir entry */

	/* get flag word first, might be too close to end of dir seg for an */
	/* entire dir entry, if this is a one-word end-of-segment marker */
	if(dirwords(d,d->buf,1)<0)
		goto flush;

	if(d->buf[0]&04000) {	/* end of dir segment, link to next */
		/* end of directory if link word was 0 */
		if(d->nextseg==0) return;

		/* make sure next seg # is in range */
		if(d->nextseg>31)
			goto flush;

		d->dirseg=d->nextseg;  /* follow link */
		d->offset=0;	/* start that segment from scratch */
		goto next;
	}

	/* get the rest of the 7-word dir entry */
	if(dirwords(d,d->buf+1,7-1)<0)
		goto flush;

	d->offset+=d->extra;	/* skip extra bytes */
	d->data+=d->size;	/* skip size of prev file to pt at this */
	d->size=d->buf[4];	/* get size of this one */

	return(0);		/* happy return */

flush:	return(-1);
}

/* flush block buffer after directory search */
static void rt11dirfinish(struct rt11dirbuf *d)
{
	if(d->bh!=NULL) brelse(d->bh);
	/* no point in setting it to NULL, they'll call rt11dirinit before */
	/* the next search starts */
}

/* read directory entries */
/* we keep reading them until end or (*filldir)() returns an error */
/* the only purpose of this routine is to gather the names of all existing */
/* files -- if Linux wants to know more, it will call rt11_read_inode() */
static int rt11_readdir(struct file *filp,void *dirent,filldir_t filldir)
{
	struct inode *i=filp->f_dentry->d_inode;  /* inode of dir */
	struct rt11dirbuf d;	/* directory buffer descriptor */
	int stored=0;		/* # entries read so far */
	char name[sizeof("FILNAM.EXT")+1];  /* +1 for NUL at end */
	char *np;		/* name pointer */
	int n;			/* # chars in filename (not including NUL) */
	off_t offset, curroff;	/* fake offset within dir */
	unsigned int ino;	/* inode # */

	offset=filp->f_pos;	/* starting position to skip to */
	curroff=0;		/* fake offset of current file entry */

	rt11dirinit(&d,i->i_dev,i->i_ino);  /* init for dir search */

	/* fetch dir entries until someone stops us */
	while(rt11dirent(&d)==0) {
		/* we have a file entry, see if we care */
		/* (permanent file and .GE. our target offset */
		if((d.buf[0]&02000)&&curroff>=offset) {
			ino=d.data;	/* inode # = blk # */

			np=name, n=0;	/* init filename */
			rad2asc(d.buf[1],&np,&n);
			rad2asc(d.buf[2],&np,&n);
			if(d.buf[3]!=0) {  /* non-null .EXT */
				*np++='.', n++;
				rad2asc(d.buf[3],&np,&n);
			}
			*np='\0';	/* mark end */

			/* copy into dirent buf */
			if((*filldir)(dirent,name,n,curroff,ino,DT_REG)<0) {
				/* failed, un-get this entry */
				d.offset-=(7*2+d.extra);
				goto flush;
			}

			/* count it, update fake file position */
			stored++;
			filp->f_pos=offset=
				curroff+sizeof("FILNAM.DAT")+1;
		}

		/* update fake directory offset (11.*entry #) */
		curroff+=sizeof("FILNAM.DAT")+1;
	}

flush:	rt11dirfinish(&d);	/* flush dir segment buffer */
	return(stored);		/* return # entries read */
}

/* check home block to make sure it looks familiar */
static int checkhomeblock(kdev_t d,unsigned int block)
{
	struct buffer_head *bh;	/* buffer descriptor for home block */
	int rc=-1;		/* return code -- assume failure */
	unsigned char *p;

	/* read home block (always block 1) */
	bh=bread(d,block,RTBLKSIZE);
	if(bh==NULL) {
		printk("rt11fs:  read error on home block\n");
		goto err2;
	}

	/* check for RT-11 signature */
	/* VMS EXCHANGE uses a different signature just to be obnoxious */
	/* (thanks to Tim Shoppa for this information) */
	p=bh->b_data;
	if(strncmp(p+0760,"DECRT11A    ",12)!=0&&
	   strncmp(p+0760,"DECVMSEXCHNG",12)!=0)
		goto err1;

	rc=0;			/* looks OK */

err1:	brelse(bh);
err2:	return(rc);
}

/* look up inode based on dir name in dentry */
static struct dentry *rt11_lookup(struct inode *dir,struct dentry *dentry)
{
	int rc;			/* return code */
	struct inode *inode;	/* the inode */
	struct rt11dirbuf d;	/* directory buffer descriptor */
	char *p;		/* pointer to filename */
	int n;			/* # chars */
	word fil, nam, ext;	/* 3-word packed RAD50 filename */
	int period=0;		/* NZ => explicit period */

	/* parse filename, convert to RAD50 */
	p=(char *)dentry->d_name.name;  /* get filename addr/length */
	n=dentry->d_name.len;
	fil=nam=ext=0;		/* init filename */
	fil=asc2rad(&p,&n);	/* parse FILNAM */
	nam=asc2rad(&p,&n);
	if(n!=0&&*p=='.') {	/* has extension */
		p++, n--;	/* eat the dot */
		ext=asc2rad(&p,&n);  /* parse EXT */
		period++;	/* remember it's explicit in case null .EXT */
	}
	rc=-EINVAL;		/* error if not all of filename taken */
	if(n!=0) goto punt;

	rt11dirinit(&d,dir->i_dev,dir->i_ino);  /* init for dir search */

	/* search dir for match */
	while(rt11dirent(&d)==0) {
		if((d.buf[0]&02000)&&	/* permanent file */
			d.buf[1]==fil&&d.buf[2]==nam&&  /* filename matches */
			(d.buf[3]==ext||  /* extension matches *OR* */
			/* ext=.DSK, user said nothing, looks like a .DSK */
			(d.buf[3]==016003&&ext==0&&!period&&
			d.size>=8&&checkhomeblock(dir->i_dev,d.data+1)==0))) {
			/* this is it */
			rt11dirfinish(&d);
			if((inode=iget(dir->i_sb,d.data))!=NULL)
				goto done;
			break;
		}
	}

	rc=-ENOENT;		/* failed */
	rt11dirfinish(&d);
	goto punt;

done:	rc=0;			/* happy */
	d_add(dentry,inode);
punt:	return(ERR_PTR(rc));
}

static int rt11_open(struct inode *i,struct file *filp)
{
	printk("<1>rt11fs:  open\n");

	return(-ENOENT);
}

static int rt11_get_block(struct inode *i,long iblock,struct buffer_head *bh,
	int create)
{
	/* fill in device, abs block #, update state */
	bh->b_dev=i->i_dev;
	bh->b_blocknr=i->i_ino+iblock;
	bh->b_state |= (1UL << BH_Mapped);
	if(create) bh->b_state |= (1UL << BH_New);
	return(0);
}

static int rt11_readpage(struct file *filp,struct page *page)
{
	return(block_read_full_page(page,rt11_get_block));
}

static struct address_space_operations rt11_aops = {
	readpage:	rt11_readpage,
};

static struct file_operations rt11_file_operations = {
/*	open:		rt11_open, */
	read:		generic_file_read,
	write:		generic_file_write,
	mmap:		generic_file_mmap,
/*	fsync:		rt11_fsync, */
};

static struct inode_operations rt11_file_inode_operations = {
/*	truncate:	rt11_truncate, */
};

static struct file_operations rt11_dir_operations = {
	read:		generic_read_dir,
	readdir:	rt11_readdir,
};

static struct inode_operations rt11_dir_inode_operations = {
	lookup:		rt11_lookup,
};

/* translate an RT-11 time to seconds since midnight UTC 01-Jan-1970 */
static time_t rt11totimet(word date)
{
	unsigned int month, day, year;

	/* split out fields: yyMMMMDDDDDYYYYY */
	/* yy'YYYYY combine to form 7-bit year since 1972 */
	/* MMMM and DDDDD are what they look like, 1-12 and 1-31 */
	month=(date>>10)&017;
	day=(date>>5)&037;
	year=((date>>(14-5))&0140)|(date&037)+1972;

	/* figure out # days since 01-Jan-1970 */
	/* algorithm from CACM, as quoted by Jim Cook on alt.sys.pdp11 */
	/* (I think this is something to do with Zoeller's Congruence) */
	if(month>2) month-=3;
	else month+=9, year--;

	return((time_t)((146097L*(year/100))/4L+
		(1461L*(year%100))/4L+
		(153L*month+2L)/5L+day-719469L)*86400L+
		sys_tz.tz_minuteswest*60);
}

/* count the number of trailing NULs in the last block of a file */
/* it would be nicer to squish out *all* NULs, but that'll burn us if it */
/* isn't a text file after all, and 99% of text files will have NULs only */
/* in the final block anyway */
static int nulls(struct inode *i,unsigned int block)
{
	struct buffer_head *bh;	/* buf descriptor for reading the block */
	int n=0;		/* # NULs counted so far */
	unsigned char *p;	/* pointer into block buf */

	/* read the block */
	bh=bread(i->i_dev,block,RTBLKSIZE);

	/* if read succeeded, count NULs */
	if(bh!=NULL) {
		for(p=bh->b_data+RTBLKSIZE;n<512;) {
			if(*--p!='\000') break;  /* finished if not NUL */
			n++;	/* count it */
		}

		brelse(bh);	/* flush the buffer */
	}

	return(n);
}

/* read a file's inode information (i.e. dir entry except for filename) */
static void rt11_read_inode(struct inode *i)
{
	struct rt11dirbuf d;	/* directory buffer descriptor */
	int prev=0;		/* starting blk # of prev file, in case it's */
				/* a logical disk containing our file */

	/* fill in file information */
	i->i_nlink=1;		/* no multi-linked files on RT-11 */
	i->i_uid=i->i_gid=0;	/* no concept of file ownership */

	/* we call the disk as a whole inode #0 */
	/* this is consistent with subdirectories (logical disks), where we */
	/* give the starting block of the .DSK file as the inode # */
	/* (so the dir always starts at inode+6) */
	if(i->i_ino==0) {
		/* must be the directory */
		i->i_mode=S_IFDIR|0644;
		i->i_size=512;	/* made-up size, meaningless on RT-11 */
		i->i_atime=i->i_ctime=i->i_mtime=0;

		/* set pointers to various directory operations */
		i->i_op=&rt11_dir_inode_operations;
		i->i_fop=&rt11_dir_operations;
		return;
	}

	/* need to search the directory for our file */
	/* each time we pass our file, we move into the subdisk that */
	/* contains it and start the search over, so we can handle an */
	/* arbitrary number of nested subdisks */
	for(;;) {
		rt11dirinit(&d,i->i_dev,prev);  /* init for dir search */

		/* search dir for this inode # */
		prev=-1;		/* no previous file yet */
		while(rt11dirent(&d)==0) {
			/* if inode # matches starting blk #, this is it */
			if(d.data==i->i_ino) {
				/* found it, grab file information */
				i->i_mode=S_IFREG|0644;
				if(d.buf[0]&0100000)	/* write-protected */
					i->i_mode&=~0200;

				/* get size in bytes */
				i->i_size=d.size*RTBLKSIZE;

				/* set all file times to the creation date */
				i->i_atime=i->i_ctime=i->i_mtime=
					rt11totimet(d.buf[6]);

				/* see if it looks like a logical disk */
				if(d.buf[3]==016003&&	/* .RAD50 "DSK" */
					d.size>=8&&	/* dir would fit */
					checkhomeblock(i->i_dev,d.data+1)==0) {

					/* yes, say it's a directory */
					i->i_mode+=S_IFDIR-S_IFREG;
					i->i_op=&rt11_dir_inode_operations;
					i->i_fop=&rt11_dir_operations;
					rt11dirfinish(&d);
					return;
				}

				/* trim trailing NULs from last block */
				/* (fixes up text files, binary files may */
				/* have wanted the NULs but they're still */
				/* there on the disk) */
				if(d.size!=0)
					i->i_size-=nulls(i,d.data+d.size-1);

				/* set pointers to various file operations */
				i->i_op=&rt11_file_inode_operations;
				i->i_fop=&rt11_file_operations;
				i->i_data.a_ops=&rt11_aops;

				rt11dirfinish(&d);
				return;
			}
			/* if we passed our file already, it must have been */
			/* inside preceding subdisk, drop out and search that */
			else if(d.data>i->i_ino) break;
			else prev=d.data; /* remember this file in case .DSK */
		}
		/* never had a hit, it must have been inside a logical disk */
		/* we broke out as soon as we hit a file which followed */
		/* ours, but if the subdisk was the last thing then we just */
		/* ran off the end of the loop, either way PREV is correct */
		rt11dirfinish(&d);	/* finish up this dir */
		if(prev<0) {
			printk("rt11fs:  never found inode %d\n",i->i_ino);
			break;
		}
	}

	/* hmm, what if anything does this do?  I saw someone do it */
/*	i->i_blocks=i->i_blksize=0; */
}

static void rt11_write_inode(struct inode *i,int wait)
{
//	lock_kernel();
//	printk("<1>rt11fs:  write_inode\n");
//	unlock_kernel();
}

static void rt11_put_inode(struct inode *i)
{
//	printk("<1>rt11fs:  put_inode\n");
}

static void rt11_delete_inode(struct inode *i)
{
	printk("<1>rt11fs:  delete_inode\n");
}

static void rt11_put_super(struct super_block *sb)
{
	printk("<1>rt11fs:  put_super\n");
}

static void rt11_write_super(struct super_block *sb)
{
	printk("<1>rt11fs:  write_super\n");

	sb->s_dirt=0;	/* saw this in some other driver, important? */
}

/* get information about the file system itself */
static int rt11_statfs(struct super_block *sb,struct statfs *sf)
{
	printk("<1>rt11fs:  statfs\n");

	sf->f_type=RTMAGIC;
	sf->f_bsize=RTBLKSIZE;
	sf->f_bfree=0;
	sf->f_bavail=0;
	sf->f_ffree=0;
/*	sf->f_blocks=...; */
	sf->f_namelen=sizeof("FILNAM.EXT");
	return(0);
}

static int rt11_remount(struct super_block *sb,int *i,char *p)
{
	printk("<1>rt11fs:  remount\n");
}

/* superblock operations */
static struct super_operations rt11_sops = {
	read_inode:	rt11_read_inode,
	write_inode:	rt11_write_inode,
	put_inode:	rt11_put_inode,
	delete_inode:	rt11_delete_inode,
	put_super:	rt11_put_super,
	write_super:	rt11_write_super,
	statfs:		rt11_statfs,
	remount_fs:	rt11_remount,
};

/* handle mounts */
static struct super_block *rt11_read_super(struct super_block *sb,
	void *options,int silent)
{
	/* set up RT-11 block size (always 512 bytes) */
	set_blocksize(sb->s_dev,RTBLKSIZE);
	sb->s_blocksize=RTBLKSIZE;
	sb->s_blocksize_bits=RTBLKBITS;

	/* check home block (always block 1) */
	if(checkhomeblock(sb->s_dev,1)<0)
		goto err;

	sb->s_magic=RTMAGIC;

	sb->s_op=&rt11_sops;		/* ptrs to handler routines */

	/* allocate root -- what does that even mean? */
	/* 2nd arg to iget() is the inode number for the root */
	if((sb->s_root=d_alloc_root(iget(sb,0)))==NULL) {
		printk("rt11fs:  failed to allocate root\n");
		goto err;
	}
/// should check for NULL?

	goto done;

err:	sb=NULL;			/* fail */
done:	return(sb);
}

/* file system type information block */
static DECLARE_FSTYPE_DEV(rt11_fs_type,"rt11",rt11_read_super);

/* add us to filesystem table */
static int __init init_rt11_fs(void)
{
	return(register_filesystem(&rt11_fs_type));
}

/* remove us on exit */
static void __exit exit_rt11_fs(void)
{
	unregister_filesystem(&rt11_fs_type);
}

EXPORT_NO_SYMBOLS;

module_init(init_rt11_fs)
module_exit(exit_rt11_fs)
